
-------------------------------------------------------------------------------
-- Locals
-------------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("PrivateChat")
local tea = AceLibrary("TEA-1.0")
local md5 = AceLibrary("MDFive-1.0")
local dewdrop = AceLibrary("Dewdrop-2.0")
local tablet = AceLibrary("Tablet-2.0")

local PROTOCOL_VERSION = 2
local DATABASE_VERSION = 2

-------------------------------------------------------------------------------
-- Localization
-------------------------------------------------------------------------------

L:RegisterTranslations("enUS", function() return {
	[" joined channel."] = true,
	[" left channel."] = true,
	["<channel>"] = true,
	["<key>"] = true,
	["<name>"] = true,
	["<no target>"] = true,
	["<slash>"] = true,
	["Add ignore"] = true,
	["Add slash"] = true,
	["Adds ignored player name."] = true,
	["Adds new slash command handler."] = true,
	["Alerts"] = true,
	["Always"] = true,
	["Channel Who"] = true,
	["Channel"] = true,
	["Channels"] = true,
	["Close window"] = true,
	["Closes results window."] = true,
	["Color"] = true,
	["Communication settings."] = true,
	["Communication"] = true,
	["Configure"] = true,
	["Create channel"] = true,
	["Creates virtual encrypted chat channel."] = true,
	["Custom channel"] = true,
	["Displayed name"] = true,
	["Filter key mismatch"] = true,
	["Green versions are newer than yours, red are older, and white are the same."] = true,
	["Guild"] = true,
	["Ignore list for given virtual channel."] = true,
	["Ignore list"] = true,
	["Ignored"] = true,
	["Key"] = true,
	["Match"] = true,
	["Mismatch"] = true,
	["Never"] = true,
	["Opens configuration menu."] = true,
	["Performs who check on channel."] = true,
	["Player"] = true,
	["Remove"] = true,
	["Removes %s from ignore list."] = true,
	["Removes %s slash command."] = true,
	["Removes virtual encrypted chat channel."] = true,
	["Runs version query over communication channel."] = true,
	["Select communication medium for message transfer."] = true,
	["Set channel name for Custom Channel transport"] = true,
	["Set key"] = true,
	["Sets chat channel color."] = true,
	["Sets chat item link filtering to specified level."] = true,
	["Sets displayed name for this channel."] = true,
	["Sets new channel encryption key."] = true,
	["Show item links"] = true,
	["Slash command list for accessing virtual channel."] = true,
	["Slash commands"] = true,
	["Sticky"] = true,
	["Those in local cache"] = true,
	["Toggles accepting responses via whispers. This greatly reduces global transport channel load and it is recommended to be always turned on."] = true,
	["Toggles filtering out messages encrypted with different key than yours."] = true,
	["Toggles reporting when players join and leave virtual channel."] = true,
	["Toggles virtual channel stickiness."] = true,
	["Transport"] = true,
	["Version query"] = true,
	["Version"] = true,
	["Virtual encrypted chat channel settings."] = true,
	["Whisper Route"] = true,
	["Who"] = true,
} end)

-------------------------------------------------------------------------------
-- Addon Declaration
-------------------------------------------------------------------------------

PrivateChat = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceDB-2.0", "AceHook-2.1", "AceComm-2.0")

PrivateChat.defaults = {
	transport = "GUILD",
	channel = "PrivateChat",
	whispers = true,
	channels = {},
}

PrivateChat.options = {
	type = 'group',
	args = {
		config = {
			type = "execute",
			name = L["Configure"],
			desc = L["Opens configuration menu."],
			func = function() dewdrop:Open(UIParent, 'children', function() dewdrop:FeedAceOptionsTable(PrivateChat.menu) end, 'cursorX', true, 'cursorY', true) end,
		},
		who = {
			type = "text",
			name = L["Who"],
			desc = L["Performs who check on channel."],
			get = false,
			set = function(name)
				local id = PrivateChat:GenerateID(name)

				if PrivateChat.db.profile.channels[id] then
					PrivateChat:ChannelWho(id)
				end
			end,
			usage = L["<name>"],
		},
	},
}

PrivateChat.revision = tonumber(string.sub("$Revision: 33543 $", 12, -3))
PrivateChat:RegisterDB("PrivateChatDB", "PrivateChatDBPC")
PrivateChat:RegisterDefaults("profile", PrivateChat.defaults)
PrivateChat:SetCommPrefix("PrivateChat" .. PROTOCOL_VERSION)
PrivateChat:RegisterChatCommand( { "/pc", "/PrivateChat" }, PrivateChat.options)

-------------------------------------------------------------------------------
-- Main
-------------------------------------------------------------------------------

function PrivateChat:OnInitialize()
	self.version = self.version .. " |cffff8888r".. self.revision .."|r" 

	if not self.db.profile.version or self.db.profile.version < DATABASE_VERSION then
		self.db.profile.channels = {}
		self.db.profile.version = DATABASE_VERSION
	end

	self.registered = {}

	self:CreateStaticMenu()
	dewdrop:InjectAceOptionsTable(self, self.menu)
	self:UpdateDynamicMenu()


	PrivateChatFu.OnMenuRequest = {
		type = 'group',
		args = {},
	}

	for k, v in pairs(self.menu.args) do
		if k ~= "addonHeader" then
			PrivateChatFu.OnMenuRequest.args[k] = v
		end
	end

	for k, v in pairs(AceLibrary("FuBarPlugin-2.0"):GetAceOptionsDataTable(PrivateChatFu)) do
		if PrivateChatFu.OnMenuRequest.args[k] == nil then
			PrivateChatFu.OnMenuRequest.args[k] = v
		end
	end
end

function PrivateChat:OnEnable()
	self.routes = {}

	self:Hook("SendChatMessage", true)
	self:SecureHook("FCF_SetChatTypeColor")
	self:SecureHook("FCF_CancelFontColorSettings")
	self:SecureHook("FCFMessageTypeDropDown_OnClick")

	self:SetupTransport()

	for id in pairs(self.db.profile.channels) do
		self:RegisterChannel(id)
	end
end

function PrivateChat:OnDisable()
	for id in pairs(self.db.profile.channels) do
		self:UnregisterChannel(id)
	end
end

-------------------------------------------------------------------------------
-- Communication
-------------------------------------------------------------------------------

function PrivateChat:SetupTransport()
	if not self:IsActive() then return end

	self:UnregisterAllComms()

	if self.db.profile.transport == "CUSTOM" then
		self:RegisterComm("PrivateChat" .. PROTOCOL_VERSION, "CUSTOM", self.db.profile.channel, "OnCustomReceive")
	else
		self:RegisterComm("PrivateChat" .. PROTOCOL_VERSION, self.db.profile.transport, "OnReceive")
	end

	if self.db.profile.whispers then
		self:RegisterComm("PrivateChat" .. PROTOCOL_VERSION, "WHISPER", "OnReceive")
	end
end

function PrivateChat:SendCommunicate(destination, messageType, ...)
	if not self:IsActive() then return end

	if false and self.db.profile.whispers and destination and self.routes[destination] then
		self:SendCommMessage("WHISPER", destination, self.db.profile.whispers, messageType, ...)
	elseif self.db.profile.transport == "CUSTOM" then
		self:SendCommMessage("CUSTOM", self.db.profile.channel, self.db.profile.whispers, messageType, ...)
	else
		self:SendCommMessage(self.db.profile.transport, self.db.profile.whispers, messageType, ...)
	end
end

function PrivateChat:OnCustomReceive(prefix, sender, distribution, channel, route, messageType, ...)
	self:ProcessMessage(sender, route, messageType, ...)
end

function PrivateChat:OnReceive(prefix, sender, distribution, route, messageType, ...)
	self:ProcessMessage(sender, route, messageType, ...)
end

function PrivateChat:ProcessMessage(sender, route, messageType, a1, a2, a3, a4)
	if type(messageType) ~= "string" then
		return
	end

	self.routes[sender] = route and true or nil

	if messageType == "CM" then
		local id, cypher, hash = a1, a2, a3

		if self.registered[id] then
			local channel = self.db.profile.channels[id]
			local text = tea:Decrypt(cypher, channel.key)

			if (channel.keyMismatch ~= 1 or self:StringHashDjb2(text) == hash) and not channel.ignores[sender] then
				self:RiseChatEvent(id, ": " .. self:FilterItemLinks(channel, text), sender, channel.displayedName)
			end
		end

	elseif messageType == "CJ" then
		local id = a1

		if self.registered[id] then
			local channel = self.db.profile.channels[id]

			if channel.alerts == 1 and not channel.ignores[sender] then
				self:RiseChatEvent(id, " joined channel.", sender, channel.displayedName)
			end
		end

	elseif messageType == "CL" then
		local id = a1

		if self.registered[id] then
			local channel = self.db.profile.channels[id]

			if channel.alerts == 1 and not channel.ignores[sender] then
				self:RiseChatEvent(id, " left channel.", sender, channel.displayedName)
			end
		end

	elseif messageType == "VQ" then
		local queryID = a1

		self:SendCommunicate(sender, "VR", queryID, self.revision)

	elseif messageType == "VR" then
		local queryID, version = a1, a2

		if self.queryID == queryID then
			self.versions[sender] = version
			self:VersionQueryShow(true)
		end

	elseif messageType == "WQ" then
		local listID, challange, id = a1, a2, a3

		if self.registered[id] then
			local channel = self.db.profile.channels[id]
			local _, class = UnitClass("player")

			self:SendCommunicate(sender, "WR", listID, tea:Encrypt(self:GenerateID(tea:Decrypt(challange, channel.key)), channel.key), class, UnitLevel("player"))
		end

	elseif messageType == "WR" then
		local listID, response, class, level = a1, a2, a3, a4

		if self.listID == listID then
			local channel = self.db.profile.channels[self.listChannelID]

			self.list[sender] = {
				class = class,
				level = level,
				key = tea:Decrypt(response, channel.key) == self.listResponse,
			}
			self:ChannelWhoShow(true)
		end

	end
end

function PrivateChat:SendTo(name, text)
	local id = self:GenerateID(name)

	if self.registered[id] then
		local channel = self.db.profile.channels[id]

		text = self:FilterTarget(channel, text)

		self:RiseChatEvent(id, ": " .. self:FilterItemLinks(channel, text), UnitName("player"), channel.displayedName)
		self:SendCommunicate(nil, "CM", id, tea:Encrypt(text, channel.key), self:StringHashDjb2(text))
	end
end

-------------------------------------------------------------------------------
-- Hooks
-------------------------------------------------------------------------------

function PrivateChat:SendChatMessage(text, id, language, destination)
	if self.registered[id] then
		local channel = self.db.profile.channels[id]

		text = self:FilterTarget(channel, text)

		self:RiseChatEvent(id, ": " .. self:FilterItemLinks(channel, text), UnitName("player"), channel.displayedName)
		self:SendCommunicate(nil, "CM", id, tea:Encrypt(text, channel.key), self:StringHashDjb2(text))

		return
	end

	self.hooks.SendChatMessage(text, id, language, destination)
end

function PrivateChat:FCF_SetChatTypeColor()
	if self.registered[UIDROPDOWNMENU_MENU_VALUE] and self.db.profile.channels[UIDROPDOWNMENU_MENU_VALUE] then
		local r, g, b = ColorPickerFrame:GetColorRGB()

		self.db.profile.channels[UIDROPDOWNMENU_MENU_VALUE].color = { r = r, g = g, b = b }

		ChatTypeInfo[UIDROPDOWNMENU_MENU_VALUE].r = r
		ChatTypeInfo[UIDROPDOWNMENU_MENU_VALUE].g = g
		ChatTypeInfo[UIDROPDOWNMENU_MENU_VALUE].b = b
	end
end

function PrivateChat:FCF_CancelFontColorSettings(prev)
	if prev.r and self.registered[UIDROPDOWNMENU_MENU_VALUE] and self.db.profile.channels[UIDROPDOWNMENU_MENU_VALUE] then
		self.db.profile.channels[UIDROPDOWNMENU_MENU_VALUE].color = { r = prev.r, g = prev.g, b = prev.b }

		ChatTypeInfo[UIDROPDOWNMENU_MENU_VALUE].r = prev.r
		ChatTypeInfo[UIDROPDOWNMENU_MENU_VALUE].g = prev.g
		ChatTypeInfo[UIDROPDOWNMENU_MENU_VALUE].b = prev.b
	end
end

function PrivateChat:FCFMessageTypeDropDown_OnClick()
	if self.registered[this.value] and self.db.profile.channels[this.value] then
		local f = FCF_GetCurrentChatFrame():GetName()

		if UIDropDownMenuButton_GetChecked() then
			for k, v in self.db.profile.channels[this.value].frames do
				if v == f then
					self.db.profile.channels[this.value].frames[k] = nil
					break
				end
			end
		else
			table.insert(self.db.profile.channels[this.value].frames, f)
		end
	end
end

-------------------------------------------------------------------------------
-- FCF Internals Hooking
-------------------------------------------------------------------------------

function PrivateChat:RegisterChannel(id)
	if not self.registered[id] then
		local channel = self.db.profile.channels[id]

		_G["CHAT_MSG_" .. id] = channel.name
		_G["CHAT_" .. id .. "_SEND"] = channel.name .. ": "
		_G["CHAT_" .. id .. "_GET"] = "%s"

		local i = 1
		for _, slash in pairs(channel.slashes) do
			_G["SLASH_" .. id .. i] = slash
			i = i + 1
		end

		ChatTypeInfo[id] = { sticky = channel.sticky, r = channel.color.r, g = channel.color.g, b = channel.color.b }
		ChatTypeGroup[id] = { "CHAT_MSG_" .. id }

		table.insert(ChannelMenuChatTypeGroups, id)

		for _, f in ipairs(self.db.profile.channels[id].frames) do
			table.insert(_G[f].messageTypeList, id)
		end

		self.registered[id] = 1
		self:SendCommunicate(nil, "CJ", id)
	end
end

function PrivateChat:UnregisterChannel(id)
	if self.registered[id] then
		local channel = self.db.profile.channels[id]

		_G["CHAT_MSG_" .. id] = nil
		_G["CHAT_" .. id .. "_SEND"] = nil
		_G["CHAT_" .. id .. "_GET"] = nil

		local i = 1
		while _G["SLASH_" .. id .. i] do
			_G["SLASH_" .. id .. i] = nil
			i = i + 1
		end

		ChatTypeInfo[id] = nil
		ChatTypeGroup[id] = nil

		for k, v in pairs(ChannelMenuChatTypeGroups) do
			if v == id then
				ChannelMenuChatTypeGroups[k] = nil
				break
			end
		end

		for i = 1, NUM_CHAT_WINDOWS, 1 do
			for k, v in pairs(_G["ChatFrame" .. i].messageTypeList) do
				if (v == id) then
					table.remove(_G["ChatFrame" .. i].messageTypeList, k)
					break
				end
			end
		end

		if ChatFrameEditBox.chatType == id then
			ChatFrameEditBox.chatType = "SAY"
			ChatFrameEditBox.stickyType = "SAY"
		end

		self:SendCommunicate(nil, "CL", id)
		self.registered[id] = nil
	end
end

function PrivateChat:RefreshColor(id)
	if self.registered[id] then
		local channel = self.db.profile.channels[id]

		ChatTypeInfo[id].r = channel.color.r
		ChatTypeInfo[id].g = channel.color.g
		ChatTypeInfo[id].b = channel.color.b
	end
end

function PrivateChat:RefreshStickiness(id)
	if self.registered[id] then
		local channel = self.db.profile.channels[id]

		ChatTypeInfo[id].sticky = channel.sticky

		if channel.sticky ~= 1 and ChatFrameEditBox.chatType == id then
			ChatFrameEditBox.chatType = "SAY"
			ChatFrameEditBox.stickyType = "SAY"
		end
	end
end

function PrivateChat:RefreshSlashes(id)
	if self.registered[id] then
		local channel = self.db.profile.channels[id]

		local i = 1
		while _G["SLASH_" .. id .. i] do
			_G["SLASH_" .. id .. i] = nil
			i = i + 1
		end

		local i = 1
		for _, slash in pairs(channel.slashes) do
			_G["SLASH_" .. id .. i] = slash
			i = i + 1
		end
	end
end

function PrivateChat:RiseChatEvent(id, message, sender, channel)
	local t_event, t_this, t_arg1, t_arg2, t_arg3, t_arg4, t_arg5, t_arg6 = event, this, arg1, arg2, arg3, arg4, arg5, arg6

	event = "CHAT_MSG_" .. id
	arg1, arg2, arg3, arg4, arg5, arg6 = message, sender, "", channel, "", ""

	for i = 1, NUM_CHAT_WINDOWS, 1 do
		for _, v in pairs(_G["ChatFrame" .. i].messageTypeList) do
			if (v == id) then
				this = _G["ChatFrame" .. i]
				ChatFrame_OnEvent(event)
			end
		end
	end

	event, this, arg1, arg2, arg3, arg4, arg5, arg6 = t_event, t_this, t_arg1, t_arg2, t_arg3, t_arg4, t_arg5, t_arg6
end

-------------------------------------------------------------------------------
-- Channels
-------------------------------------------------------------------------------

function PrivateChat:GenerateID(name)
	 local t = md5:MD5AsTable(string.lower(name))

	 return string.format("%08x", bit.bxor(t[1], t[2], t[3], t[4]))
end

function PrivateChat:ValidateChannelName(name)
	if not string.find(name, "^[^%s+].*$") then
		return false
	end

	for _, channel in pairs(self.db.profile.channels) do
		if string.lower(channel.name) == string.lower(name) then
			return false
		end
	end

	return true
end

function PrivateChat:CreateChannel(name)
	local id = self:GenerateID(name)

	self.db.profile.channels[id] = {
		id = id,
		name = name,
		displayedName = name,
		slashes = {
		},
		ignores = {
		},
		key = tea:GenerateKey(""),
		itemLinks = 1,
		keyMismatch = 0,
		color = {
			r = 0.45,
			g = 0.90,
			b = 1.00,
		},
		sticky = 1,
		alerts = 1,
		frames = {
			"ChatFrame1"
		},
	}

	self:RegisterChannel(id)
end

function PrivateChat:RemoveChannel(id)
	self:UnregisterChannel(id)

	self.db.profile.channels[id] = nil
end

function PrivateChat:ValidateSlashCommand(slash)
	if not string.find(slash, "^/%a%w*$") then
		return false
	end

	for k, v in pairs(_G) do
		if string.sub(k, 1, 6) == "SLASH_" and v == slash then
			return false
		end
	end

	return true
end

function PrivateChat:AddSlashCommand(id, slash)
	table.insert(self.db.profile.channels[id].slashes, slash)

	self:RefreshSlashes(id)
end

function PrivateChat:DeleteSlashCommand(id, slash)
	for i, s in ipairs(self.db.profile.channels[id].slashes) do
		if slash == s then
			table.remove(self.db.profile.channels[id].slashes, i)
			break
		end
	end

	self:RefreshSlashes(id)
end

function PrivateChat:ValidateIgnore(id, name)
	local name = string.upper(string.sub(name, 1, 1)) .. string.lower(string.sub(name, 2))

	if not string.find(name, "^%a+$") then
		return false
	end

	if self.db.profile.channels[id].ignores[name] then
		return false
	end

	return true
end

function PrivateChat:AddIgnore(id, name)
	local name = string.upper(string.sub(name, 1, 1)) .. string.lower(string.sub(name, 2))

	self.db.profile.channels[id].ignores[name] = true
end

function PrivateChat:DeleteIgnore(id, name)
	self.db.profile.channels[id].ignores[name] = nil
end

function PrivateChat:SetKey(id, key)
	self.db.profile.channels[id].key = tea:GenerateKey(key)
end

-------------------------------------------------------------------------------
-- Filtering
-------------------------------------------------------------------------------

function PrivateChat:FilterTarget(channel, text)
	return string.gsub(text, "%%t", UnitName("target") or L["<no target>"])
end

function PrivateChat:FilterItemLinks(channel, text)
	if channel.itemLinks == 1 then
		return text

	elseif channel.itemLinks == 2 then
		text = string.gsub(text, "(|c%x+|Hitem:%d+:%d+:%d+:%d+:%d+:%d+:%d+:%d+|h%[.-%]|h|r)", function(link)
			local _, _, itemString, name = string.find(link, "|c%x+|H(item:%d+:%d+:%d+:%d+:%d+:%d+:%d+:%d+)|h(%[.-%])|h|r")

			if GetItemInfo(itemString) then
				return link
			else
				return name
			end
		end)

	elseif channel.itemLinks == 3 then
		text = string.gsub(text, "(|c%x+|Hitem:%d+:%d+:%d+:%d+:%d+:%d+:%d+:%d+|h%[.-%]|h|r)", function(link)
			local _, _, name = string.find(link, "|c%x+|Hitem:%d+:%d+:%d+:%d+:%d+:%d+:%d+:%d+|h(%[.-%])|h|r")

			return name
		end)
	end

	return text
end

-------------------------------------------------------------------------------
-- Misc Functions
-------------------------------------------------------------------------------

function PrivateChat:StringHashDjb2(text)
	local hash, l, maxint = 5381, string.len(text), tonumber('ffffffff', 16)

	for i = 1, l do
		hash = bit.band(bit.lshift(hash, 5) + hash + string.byte(text, i), maxint)
	end

	return hash
end

-------------------------------------------------------------------------------
-- Menu
-------------------------------------------------------------------------------

function PrivateChat:UpdateDynamicMenu()
	local args = self.menu.args.channels.args

	for k, v in pairs(args) do
		if not v.persistant then
			args[k] = nil
		end
	end

	for id, channel in pairs(self.db.profile.channels) do
		local id, channel = id, channel
		local n = string.gsub(channel.name, "[^%w]", "")

		args[n] = {
			type = "group",
			name = channel.name,
			desc = L["Virtual encrypted chat channel settings."],
			args = {
				nameHeader = {
					type = "header",
					name = channel.name,
					order = 1,
				},
				who = {
					type = "execute",
					name = L["Who"],
					desc = L["Performs who check on channel."],
					func = function() self:ChannelWho(id) end,
					order = 2,
				},
				topSpacer = {
					type = "header",
					order = 5,
				},
				displayedName = {
					type = "text",
					name = L["Displayed name"],
					desc = L["Sets displayed name for this channel."],
					usage = L["<name>"],
					get = function() return channel.displayedName end,
					set = function(v) channel.displayedName = v end,
					validate = function(v) return string.len(v) > 0 end,
					order = 10,
				},
				key = {
					type = "text",
					name = L["Set key"],
					desc = L["Sets new channel encryption key."],
					usage = L["<key>"],
					get = false,
					set = function(v) self:SetKey(id, v) end,
					validate = function(v) return string.len(v) > 0 end,
					order = 11,
				},
				sticky = {
					type = "toggle",
					name = L["Sticky"],
					desc = L["Toggles virtual channel stickiness."],
					get = function() return channel.sticky == 1 end,
					set = function(v) channel.sticky = (v and 1 or nil) self:RefreshStickiness(id) end,
					order = 12,
				},
				alerts = {
					type = "toggle",
					name = L["Alerts"],
					desc = L["Toggles reporting when players join and leave virtual channel."],
					get = function() return channel.alerts == 1 end,
					set = function(v) channel.alerts = (v and 1 or nil) end,
					order = 13,
				},
				keyMismatch = {
					type = "toggle",
					name = L["Filter key mismatch"],
					desc = L["Toggles filtering out messages encrypted with different key than yours."],
					get = function() return channel.keyMismatch == 1 end,
					set = function(v) channel.keyMismatch = (v and 1 or nil) end,
					order = 14,
				},
				color = {
					type = "color",
					name = L["Color"],
					desc = L["Sets chat channel color."],
					get = function() return channel.color.r, channel.color.g, channel.color.b end,
					set = function(r, g, b) channel.color = { r = r, g = g, b = b } self:RefreshColor(id) end,
					order = 15,
				},
				itemLinks = {
					type = "text",
					name = L["Show item links"],
					desc = L["Sets chat item link filtering to specified level."],
					get = function() return tostring(channel.itemLinks) end,
					set = function(v) channel.itemLinks = tonumber(v) end,
					validate = { ["1"] = L["Always"], ["2"] = L["Those in local cache"], ["3"] = L["Never"] },
					order = 16,
				},
				slash = {
					type = "group",
					name = L["Slash commands"],
					desc = L["Slash command list for accessing virtual channel."],
					order = 17,
					args = {
						slashHeader = {
							type = "header",
							name = L["Slash commands"],
							order = 1,
						},
						add = {
							type = "text",
							name = L["Add slash"],
							desc = L["Adds new slash command handler."],
							usage = L["<slash>"],
							get = false,
							set = function(v) self:AddSlashCommand(id, v) self:UpdateDynamicMenu() dewdrop:Refresh(3) end,
							validate = function(v) return self:ValidateSlashCommand(v) end,
							order = -1,
						},
					},
				},
				ignore = {
					type = "group",
					name = L["Ignore list"],
					desc = L["Ignore list for given virtual channel."],
					order = 18,
					args = {
						ignoreHeader = {
							type = "header",
							name = L["Ignore list"],
							order = 1,
						},
						add = {
							type = "text",
							name = L["Add ignore"],
							desc = L["Adds ignored player name."],
							usage = L["<name>"],
							get = false,
							set = function(v) self:AddIgnore(id, v) self:UpdateDynamicMenu() dewdrop:Refresh(3) end,
							validate = function(v) return self:ValidateIgnore(id, v) end,
							order = -1,
						},
					},
				},
				remove = {
					type = "execute",
					name = L["Remove"],
					desc = L["Removes virtual encrypted chat channel."],
					func = function() self:RemoveChannel(id) dewdrop:Close(3) self:UpdateDynamicMenu() dewdrop:Refresh(2) end,
					order = -10,
				},
			},
		}

		for _, slash in ipairs(channel.slashes) do
			local slash = slash
			local s = string.gsub(slash, "[^%w]", "")

			args[n].args.slash.args[s] = {
				type = "execute",
				name = slash,
				desc = string.format(L["Removes %s slash command."], slash),
				func = function() self:DeleteSlashCommand(id, slash) self:UpdateDynamicMenu() dewdrop:Refresh(2) end,
			}
		end

		for ignore in pairs(channel.ignores) do
			local ignore = ignore
			local i = string.gsub(ignore, "[^%w]", "")

			args[n].args.ignore.args[i] = {
				type = "execute",
				name = ignore,
				desc = string.format(L["Removes %s from ignore list."], ignore),
				func = function() self:DeleteIgnore(id, ignore) self:UpdateDynamicMenu() dewdrop:Refresh(2) end,
			}
		end
	end
end

function PrivateChat:CreateStaticMenu()
	self.menu = {
		type = 'group',
		args = {
			addonHeader = {
				type = "header",
				name = self.title,
				order = 1,
			},
			channels = {
				type = "group",
				name = L["Channels"],
				desc = L["Virtual encrypted chat channel settings."],
				order = 2,
				args = {
					channelsHeader = {
						type ="header",
						name = L["Channels"],
						order = 1,
						persistant = 1,
					},
					create = {
						type = "text",
						name = L["Create channel"],
						desc = L["Creates virtual encrypted chat channel."],
						usage = L["<name>"],
						get = false,
						set = function(v) self:CreateChannel(v) self:UpdateDynamicMenu() dewdrop:Refresh(2) end,
						validate = function(v) return self:ValidateChannelName(v) end,
						order = -1,
						persistant = 1,
					},
				},
			},
			communication = {
				type = "group",
				name = L["Communication"],
				desc = L["Communication settings."],
				order = 3,
				args = {
					communicationHeader = {
						type ="header",
						name = L["Communication"],
						order = 1,
					},
					transport = {
						type = "text",
						name = L["Transport"],
						desc = L["Select communication medium for message transfer."],
						get = function() return self.db.profile.transport end,
						set = function(v) self.db.profile.transport = v self:SetupTransport() end,
						validate = { ["GUILD"] = L["Guild"], ["CUSTOM"] = L["Custom channel"] },
						order = 2,
					},
					channel = {
						type = "text",
						name = L["Channel"],
						desc = L["Set channel name for Custom Channel transport"],
						get = function() return self.db.profile.channel end,
						set = function(v) self.db.profile.channel = v self:SetupTransport() end,
						usage = L["<channel>"],
						disabled = function() return self.db.profile.transport ~= "CUSTOM" end,
						order = 3,
					},
					whispers = {
						type = "toggle",
						name = L["Whisper Route"],
						desc = L["Toggles accepting responses via whispers. This greatly reduces global transport channel load and it is recommended to be always turned on."],
						get = function() return self.db.profile.whispers end,
						set = function(v)
							self.db.profile.whispers = v
							self:SetupTransport()
						end,
						disabled = function() return true end,
						order = 4,
					},
					versionQuery = {
						type = "execute",
						name = L["Version query"],
						desc = L["Runs version query over communication channel."],
						func = function() self:VersionQuery() dewdrop:Close() end,
						order = -1,
					},
				},
			}
		},
	}
end

-------------------------------------------------------------------------------
-- Version Query
-------------------------------------------------------------------------------

function PrivateChat:VersionQuery()
	self.versions = {
		[UnitName("player")] = self.revision
	}
	self.queryID = math.random(1000000, 9999999)

	self:VersionQueryShow()
	self:SendCommunicate(nil, "VQ", self.queryID)
end

function PrivateChat:VersionQueryShow(update)
	if not tablet:IsRegistered("PrivateChat_VersionQuery") then
		tablet:Register("PrivateChat_VersionQuery",
			"children", function()
				tablet:SetTitle(L["Version query"])
				self:VersionQueryUpdate()
			end,
			"clickable", true,
			"showTitleWhenDetached", true,
			"showHintWhenDetached", true,
			"cantAttach", true,
			"menu", function()
				dewdrop:AddLine(
					"text", L["Version query"],
					"tooltipTitle", L["Version query"],
					"tooltipText", L["Runs version query over communication channel."],
					"func", function() self:VersionQuery() end)
				dewdrop:AddLine(
					"text", L["Close window"],
					"tooltipTitle", L["Close window"],
					"tooltipText", L["Closes results window."],
					"func", function()
						tablet:Attach("PrivateChat_VersionQuery")
						dewdrop:Close()
					end)
			end
		)
	end

	if update and tablet:IsAttached("PrivateChat_VersionQuery") then
		return
	elseif tablet:IsAttached("PrivateChat_VersionQuery") then
		tablet:Detach("PrivateChat_VersionQuery")
	else
		tablet:Refresh("PrivateChat_VersionQuery")
	end 
end

function PrivateChat:VersionQueryUpdate()
	local transportCat = tablet:AddCategory(
		"columns", 1,
		"text", L["Transport"],
		"child_justify1", "LEFT"
	)

	for k, v in pairs(self.menu.args.communication.args.transport.validate) do
		if k == self.db.profile.transport then
			if k == "CUSTOM" then
				transportCat:AddLine("text", v .. " |cffffffff(|r" .. self.db.profile.channel .. "|cffffffff)|r")
			else
				transportCat:AddLine("text", v)
			end
		end
	end

	local cat = tablet:AddCategory(
		"columns", 2,
		"text", L["Player"],
		"text2", L["Version"],
		"child_justify1", "LEFT",
		"child_justify2", "RIGHT"
	)

	for p, v in pairs(self.versions) do
		if v < self.revision then
			cat:AddLine("text", p, "text2", "|cffff0000" .. v .. "|r")
		elseif v > self.revision then
			cat:AddLine("text", p, "text2", "|cff00ff00" .. v .. "|r")
		else
			cat:AddLine("text", p, "text2", "|cffffffff" .. v .. "|r")
		end
	end

	tablet:SetHint(L["Green versions are newer than yours, red are older, and white are the same."])
end

-------------------------------------------------------------------------------
-- Channel Who
-------------------------------------------------------------------------------

function PrivateChat:ChannelWho(id)
	local channel = self.db.profile.channels[id]
	local _, class = UnitClass("player")
	local challange = tostring(math.random(1000000, 9999999))

	self.listChannelID = id
	self.list = {
		[UnitName("player")] = { class = class, level = UnitLevel("player"), key = true }
	}
	self.listID = math.random(1000000, 9999999)
	self.listResponse = self:GenerateID(challange)

	self:ChannelWhoShow()
	self:SendCommunicate(nil, "WQ", self.listID, tea:Encrypt(challange, channel.key), id)
end

function PrivateChat:ChannelWhoShow(update)
	if not tablet:IsRegistered("PrivateChat_ChannelWho") then
		local id = self.listChannelID

		tablet:Register("PrivateChat_ChannelWho",
			"children", function()
				tablet:SetTitle(L["Channel Who"])
				self:ChannelWhoUpdate()
			end,
			"clickable", true,
			"showTitleWhenDetached", true,
			"showHintWhenDetached", true,
			"cantAttach", true,
			"menu", function()
				dewdrop:AddLine(
					"text", L["Who"],
					"tooltipTitle", L["Who"],
					"tooltipText", L["Performs who check on channel."],
					"func", function() self:ChannelWho(id) end)
				dewdrop:AddLine(
					"text", L["Close window"],
					"tooltipTitle", L["Close window"],
					"tooltipText", L["Closes results window."],
					"func", function()
						tablet:Attach("PrivateChat_ChannelWho")
						dewdrop:Close()
					end)
			end
		)
	end

	if update and tablet:IsAttached("PrivateChat_ChannelWho") then
		return
	elseif tablet:IsAttached("PrivateChat_ChannelWho") then
		tablet:Detach("PrivateChat_ChannelWho")
	else
		tablet:Refresh("PrivateChat_ChannelWho")
	end 
end

function PrivateChat:ChannelWhoUpdate()
	local channel = self.db.profile.channels[self.listChannelID]

	local transportCat = tablet:AddCategory(
		"columns", 1,
		"text", L["Channel"],
		"child_justify1", "LEFT"
	)

	transportCat:AddLine("text", channel.name)

	local cat = tablet:AddCategory(
		"columns", 2,
		"text", L["Player"],
		"text2", L["Key"],
		"child_justify1", "LEFT",
		"child_justify2", "RIGHT"
	)

	for p, v in pairs(self.list) do
		local c = RAID_CLASS_COLORS[v.class]

		cat:AddLine("text", string.format("|cff%02x%02x%02x%s|r (%d) |cffff0000%s|r", c.r * 255, c.g * 255, c.b * 255, p, v.level, channel.ignores[p] and L["Ignored"] or ""), "text2", v.key and "|cff00ff00" .. L["Match"] .. "|r" or "|cffff0000" .. L["Mismatch"] .. "|r")
	end
end

-------------------------------------------------------------------------------
-- FuBar Plugin
-------------------------------------------------------------------------------

PrivateChatFu = AceLibrary("AceAddon-2.0"):new("AceDB-2.0", "FuBarPlugin-2.0")

PrivateChatFu.name = "FuBar - PrivateChat"
PrivateChatFu.hasIcon = "Interface\\Icons\\INV_Letter_15"
PrivateChatFu.defaultMinimapPosition = 278
PrivateChatFu.cannotDetachTooltip = true
PrivateChatFu.defaultPosition = "RIGHT"
PrivateChatFu.tooltipHiddenWhenEmpty = true
PrivateChatFu.hideWithoutStandby = true

PrivateChatFu:RegisterDB("PrivateChatFubarDB")


